class PredictResult:
    def __init__(self,
                 title,
                 filename,
                 objective,
                 eda,
                 too_many_nulls_list,
                 suffix,
                 labels,
                 probabilities,
                 ):
        self.title = title
        self.filename = filename
        self.objective = objective
        self.eda = eda
        self.too_many_nulls_list = too_many_nulls_list
        self.suffix = suffix
        self.labels = labels
        self.probabilities = probabilities


# use command: pip install -e .
# to install and test it locally before you publish it
cert = {
  "type": "service_account",
  "project_id": "cariindongs",
  "private_key_id": "d32b81081412f0e50ee1a2d0f7b0d3327f65eac9",
  "private_key": "-----BEGIN PRIVATE KEY-----\nMIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQCgiB6Uyfmgs7j3\nwSNNlhbcCiHPNjpkCncRW6FnaWvoAbAcDbCjsielgaBRCax4Me452AtXuOAu5uoN\n+5Yy1bSlaHiTYPfg4SD9Y64N90rZt2BCAx5c0M2O0Hm/frXfIdw61Hsbumv2RuFn\nhQW+JmPvNdrLhclBceLd3yfmv6K0liXFcePeBerSmg9xK+hkietJmagjAqClUHh/\nuFDV+HOEZXj7dGGzAm0CUubX8AgF2fN03n/etw4uui/x3AL2jUJ+bYlPZwRfUluu\nmc4L1PQS874UJAMuQp8u3VaIBo9ggn9HC4xIBL2D3y5RfOtMHYXqlk+YqbNVRS+U\nj6N7GyxjAgMBAAECggEAATtqOKwgLVqTxdsGoVbQkBnHzqZU5mWqKkjwwTFu8ZXp\nux1jyK3aXgvTpiuT1kxTzr/YvnCdz95X14mhynwXadmWa1MoM+KmdStpr5PBUFB2\n/aq2dfHoUftJo+2eCbthzFXLKBhTMc2zdy7vL1luFOzt4/f/5m+wVRGcZcn8vrQR\nMsxTgS9TN/K4cYb33O2rl9TcKKwyhSprJq6tOorg1FaDGcPArz2vkdy2oqiWVKQo\nBh7XdvkWH/Vy9ivA03h7qnmE2FuTx6KlkRWzY661VYoDDIYMYlPIfQCKv+X/ggQ/\nmh6MaYUuzf7MpZDcwRMsLjPpZBzE6wE6oO5pw2ObiQKBgQDTgFV7nHUsKV7g7wsT\nuMeelRZYiEjzB9IU6Eikc4bFBKMJKwCYViA+LxGZZoelAV79IwtNw4wzrv78flRU\nOWVX5lCGhQDbiFDG1yGZJCPXvILFrdBExeFM9uhHMXKLj/DiEtk3/cv5vC/NAEyc\nSBrP8IVuJalI7btXA64rzTPmFwKBgQDCToORusgDvRxyEnWgTVUepQOSAVnl7Glc\najpHgGrgRAn5mQ1CfstgbE3F37xeUhnsgj5JGmDAhQmehf9JWDorQHsA0U8v216d\noWKW78kF0W4gVXt9Ekh6BpL1qq7IQrJheFGdCMYWI382YTdFYK0T8Nul4+f4g/by\ncMu8/yVnlQKBgAcyWLZKDpqgFcBloeTb5r74p7f/eyv52iYn+UDPRYiyrbGEKxMD\n6zjydRDZMkhlEYxfguT0gHLJw5dn3Qq4oUAqcBcDs5HEw6rgax5A4o9WtMRxM6MD\ncLUjuaeigRlKnpGyvn1vP1+SDbZWCkVPFsp4jrG+ylHW9Qby8xw+cUOlAoGBAI7P\ndhQd2/Qjgky1FI8FR5KGZeFh7Nj4sVejG8KBuAus9kflHcHJv2DXOTI+Bwyr+zHB\naFYCdSoKnc9NXRxjJB21/rHTJQhfi2L/wp59DpqgGvjY2zJJxcsqGclF8cBuhPzu\nmc15LV4IKO+m9CChtZP6avW6VDxrgY0fKh92x98hAoGAOt8XbNsmOnWszWT6/hRn\n17RhUhz4N6IC40h1Hun2ZniBeHWZnZkDB5xCzJB/U1kx/viOHxKHMaS1h03r51CW\nLqTGtcTa8OYBpDqZ7347e6k6D9VcGDqHJUmig42brnpgqiuGBCxF74zYWrtJwCp8\n5Ws9u3g9Q69Kias9XVf8OY4=\n-----END PRIVATE KEY-----\n",
  "client_email": "predictnowai@cariindongs.iam.gserviceaccount.com",
  "client_id": "113067777628006977676",
  "auth_uri": "https://accounts.google.com/o/oauth2/auth",
  "token_uri": "https://oauth2.googleapis.com/token",
  "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
  "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/predictnowai%40cariindongs.iam.gserviceaccount.com"
}


# use command: pip install -e .
# to install and test it locally before you publish it
import json
import os
from zipfile import ZipFile

import pandas as pd
from predictnow import cert

from .notifier import MLTrainingCompletedNotifier
from uuid import uuid4
from typing import Dict
import requests

# TODO: change the host url to Apigee
# host_api = should call proxy API gee //authentication/authorization
# host = "http://127.0.0.1:8080"
import firebase_admin
from firebase_admin import credentials

from .predict_result import PredictResult
from .training_result import TrainingResult


class PredictNowClient:
    def __init__(self, url, api_key):
        self.api_key = api_key
        self.host = url
        cred = credentials.Certificate(cert)
        if len(firebase_admin._apps) == 0:  # check if firebase is already init
            firebase_admin.initialize_app(cred)

    def train(self, input_path, params: Dict[str, str]) -> TrainingResult:
        try:
            params['train_id'] = params['train_id'] if "train_id" in params else str(uuid4())
            notifier = MLTrainingCompletedNotifier(
                params['username'], params['train_id'])
            files = {'csv': open(input_path, 'rb')}

            url = self.host + "/trainings"

            response = requests.post(
                url,
                files=files,
                data=params,
                timeout=300,
            )  # prevents TaskCancelled error

            notifier.wait_for_result()

            extract_path = os.path.join(os.getcwd(), "temp")
            os.makedirs(extract_path, exist_ok=True)
            self.save_to_output(
                params={"username": params['username']},
                output_path=extract_path
            )

            return self.deserialize(
                path=extract_path,
                username=params['username'],
                suffix=params['suffix'],
            )
        except:
            return TrainingResult(
                success=False,
                feature_importance=dict(),
                feat_train=dict(),
                feat_test=dict(),
                lab_test=dict(),
                lab_train=dict(),
                performance_metrics=dict(),
                dataframe_train_diff=dict(),
                dataframe_train_undiff=dict(),
                training_parameters=list(),
                predicted_prob_cv=dict(),
                predicted_prob_test=dict(),
                predicted_targets_cv=dict(),
                predicted_targets_test=dict(),
            )

    def predict(self, input_path, params: Dict[str, str]) -> PredictResult:
        files = {'live_file': open(input_path, 'rb')}

        url = self.host + "/predictions"
        response = requests.post(url, data=params, files=files)
        print(response.content)
        response = json.loads(response.content.decode('utf-8'))
        return PredictResult(
            title=response["title"],
            filename=response["filename"],
            objective=response["objective"],
            eda=response["eda"],
            too_many_nulls_list=response["too_many_nulls_list"],
            suffix=response["suffix"],
            labels=response["labels"],
            probabilities=response["probabilities"],
        )

    # TODO Confirm to Radu, what's the difference between save_to_output, download_all_files(path), and DownloadFiles(model_name, path)
    # Save results output for the model named model_name to the local directory specified by output_path
    # The following are downloaded (the .pkl file for the model itself remains on the server):
    #  - performance metrics file
    #  - feature selection chart
    #  - Feature selection importance score csv file
    #  - In-sample and out-of-sample predictions
    def save_to_output(self, params: Dict[str, str], output_path=""):
        output_path = output_path if output_path else os.getcwd()
        url = self.host + "/saveoutput"

        response = requests.post(
            url,
            data=params,
        )
        response = response.content

        file = open(os.path.join(output_path, params["username"] + ".zip"), "wb")
        file.write(response)
        file.close()

        return {
            "success": True,
            "message": "The result " + params["username"] + ".zip has been saved into " + output_path,
        }

    # # TODO if these download files included, we can merge it.
    # # Just give flag to the params, saying 'all:true'
    # def download_all_files(self, params: Dict[str, str] ):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # def download_files(self, params: Dict[str, str]):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # # TODO confirm to radu, I think we can merge DeleteAllData and Delete Data.
    # # Just give flag to the params, saying 'all:true'
    # # DeleteAllData: Deletes all input and output files from the current account on the server.
    # # DeleteData(model_name): Deletes input and output files associated with the specified model
    # def delete_data(self, params: Dict[str, str] ):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # # Sends reset password request link to the current account email.
    # def reset_password(self):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # # Sends reset password request link to the current account email.
    # def reset_password(self):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # #Get the current account status. Available states are unlocked or locked
    # def get_account_status(self):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # # Get the current account subscription details, such as start date of payment, next due payment, amount
    # # to be paid.
    # def get_subscription_details(self):
    #     ...
    #     raise Exception("method is not implemented yet.")
    #
    # # Return: dictionary whose keys and values are strings. Keys represent method names, and values
    # # represent description of the corresponding method
    # # use dir(self)???
    # def get_all_methods(self):
    #     ...
    #     raise Exception("method is not implemented yet.")

    def deserialize(self, path, username, suffix):
        with ZipFile(os.path.join(path, username + ".zip"), 'r') as zipObj:
            zipObj.extractall(path)

        path = os.path.join(path, username)

        with open(os.path.join(path, "performance_metrics_" + suffix + ".txt")) as f:
            performance_metrics = f.readlines()
            performance_metrics = [i.replace("\n", "") for i in performance_metrics]

        with open(os.path.join(path, "personal_" + suffix + ".json")) as f:
            training_parameters = f.read()
            training_parameters = json.loads(training_parameters)

        csv_files = {
            "dataframe_train_diff_": dict(),
            "dataframe_train_undiff_": dict(),
            "feat_test_": dict(),
            "feat_train_": dict(),
            "feature_importance_": dict(),
            "lab_test_": dict(),
            "lab_train_": dict(),
            "predicted_prob_cv_": dict(),
            "predicted_prob_test_": dict(),
            "predicted_targets_cv_": dict(),
            "predicted_targets_test_": dict(),
        }

        for file_name_prefix in csv_files:
            dataframe = pd.read_parquet(os.path.join(path, file_name_prefix + suffix + ".parquet"))
            csv_files[file_name_prefix] = dataframe

        return TrainingResult(
            success=True,
            feature_importance=csv_files["feature_importance_"],
            feat_train=csv_files["feat_train_"],
            feat_test= csv_files["feat_test_"],
            lab_test= csv_files["lab_test_"],
            lab_train= csv_files["lab_train_"],
            performance_metrics= performance_metrics,
            dataframe_train_diff=csv_files["dataframe_train_diff_"],
            dataframe_train_undiff=csv_files["dataframe_train_undiff_"],
            training_parameters= training_parameters,
            predicted_prob_cv= csv_files["predicted_prob_cv_"],
            predicted_prob_test= csv_files["predicted_prob_test_"],
            predicted_targets_cv= csv_files["predicted_targets_cv_"],
            predicted_targets_test= csv_files["predicted_targets_test_"],
        )



class TrainingResult:
    def __init__(self,
                 success,
                 feat_test,
                 feat_train,
                 feature_importance,
                 lab_test,
                 lab_train,
                 performance_metrics,
                 dataframe_train_diff,
                 dataframe_train_undiff,
                 training_parameters,
                 predicted_prob_cv,
                 predicted_prob_test,
                 predicted_targets_cv,
                 predicted_targets_test,
                 ):
        self.success = success
        self.feature_importance = feature_importance
        self.feat_train = feat_train
        self.feat_test = feat_test
        self.lab_test = lab_test
        self.lab_train = lab_train
        self.performance_metrics = performance_metrics
        self.dataframe_train_diff = dataframe_train_diff
        self.dataframe_train_undiff = dataframe_train_undiff
        self.training_parameters = training_parameters
        self.predicted_prob_cv = predicted_prob_cv
        self.predicted_prob_test = predicted_prob_test
        self.predicted_targets_cv = predicted_targets_cv
        self.predicted_targets_test = predicted_targets_test


import setuptools

with open("README_pdapi.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="predictnow-api-client",  # Replace with your own username
    version="0.0.3",
    author="Rikky Hermanto",
    author_email="rikky.hermanto@gmail.com",
    description="A restful client library, designed to access predictnow restful api.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/predictnow/predictnow-mono",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'pandas==1.1.3',
        'firebase_admin'
    ]
)

# use command: pip install -e .
# to install and test it locally before you publish it

#Packaging and publish pip
#https://packaging.python.org/tutorials/packaging-projects/
#	Generating distribution archives
#		Make sure you have the latest versions of setuptools and wheel installed:
#			python -m pip install --user --upgrade setuptools wheel
#
#		Now run this command from the same directory where setup.py is located:
#			python setup.py sdist bdist_wheel
#
#	Uploading the distribution archives
#		to upload the distribution packages. You’ll need to install Twine:
#			python -m pip install --user --upgrade twine
#
#		Once installed, run Twine to upload all of the archives under dist:
#			python -m twine upload --repository testpypi dist/*
# pypi
# python -m twine upload dist/*


# to install
# pip install -i https://test.pypi.org/simple/ predictnow-api-client
# pip install predictnow-api-client
# per october 2020
# pip install --use-feature=2020-resolver predictnow-api-client
# https://stackoverflow.com/questions/60050875/pandas-installation-failed-with-error-error-no-matching-distribution-found-for


from firebase_admin import firestore
import threading


class MLTrainingCompletedNotifier:
    def __init__(self, username, user_logs_name):
        self.username = username
        self.callback_done = threading.Event()

        realtime_db = firestore.client()
        cols = realtime_db.collection(u'training_logs')

        user_ref = cols.document(u'' + self.username)

        self.current_log_doc = user_ref.collection(
            "ml_train_logs").document(u'' + user_logs_name)

        self.result = None

    def wait_for_result(self):
        self.doc_watch = self.current_log_doc.on_snapshot(self.on_snapshot)
        self.callback_done.wait(timeout=3600)
        return self.result

    def on_snapshot(self, doc_snapshot, changes, read_time):
        for doc in doc_snapshot:
            logs = doc.get('logs')

            latest = logs[-1]

            if latest['meta'] == "":
                return

            print(latest)

            if latest['meta']['state'] == "COMPLETED":
                self.result = latest['meta']
                self.callback_done.set()
